APP_VERSION_REGISTRY_KEY = 'snovault.app_version'

# This order determines order that items will be mapped + added to the queue
# Can use item type (e.g. file_fastq) or class name (e.g. FileFastq)
# This order is not meaningful for snovault as it is not a standalone
# app. This value should be overridden in the downstream application
ITEM_INDEX_ORDER = []
